/** EMULib Emulation Library *********************************/
/**                                                         **/
/**                        SndALSA.c                        **/
/**                                                         **/
/** This file contains ALSA-dependent sound implementation  **/
/** for the emulation library.                              **/
/**                                                         **/
/** Copyright (C) Marat Fayzullin 1996-2010                 **/
/**     You are not allowed to distribute this software     **/
/**     commercially. Please, notify me, if you make any    **/
/**     changes to this file.                               **/
/*************************************************************/
#include "EMULib.h"
#include "Sound.h"

#include <ctype.h>
#include <stdio.h>
#include <sys/types.h>
#include <asm/types.h>
#include <linux/types.h>
#include <sys/ioctl.h>

#ifndef OLD_ALSA
#include <sound/asound.h>
#else
#include <alsa/asoundlib.h>
#define SNDRV_PCM_STREAM_PLAYBACK       SND_PCM_STREAM_PLAYBACK
#define SNDRV_PCM_ACCESS_RW_INTERLEAVED SND_PCM_ACCESS_RW_INTERLEAVED
#define SNDRV_PCM_FORMAT_S16_LE         SND_PCM_FORMAT_S16_LE
#define SNDRV_PCM_FORMAT_S8             SND_PCM_FORMAT_S8
#endif

static void *ALSA     = 0;  /* Handle to ALSA audio driver   */
static int SndRate;         /* ALSA sampling rate            */
static int SndSize;         /* ALSA audio buffer size        */
static int AudioPaused;     /* 1: Audio is paused            */

/** InitAudio() **********************************************/
/** Initialize sound. Returns rate (Hz) on success, else 0. **/
/** Rate=0 to skip initialization (will be silent).         **/
/*************************************************************/
unsigned int InitAudio(unsigned int Rate,unsigned int Latency)
{
  struct snd_pcm_hw_params *HWParams;
  struct snd_pcm_sw_params *SWParams;
  struct snd_output *Log;
  int J,I;

  /* Uninitialize sound, just in case */
  TrashSound();

  /* No sound yet */
  ALSA        = 0;
  AudioPaused = 0;
  SndRate     = 0;

  /* If not initializing sound, drop out */
  if(!Rate) return(0);

  /* Open ALSA PCM device */
  if((J=snd_pcm_open(&ALSA,"default",SNDRV_PCM_STREAM_PLAYBACK,0))<0) 
  {
    printf("ALSA: %s (%d)\n",snd_strerror(J),J);
    ALSA = 0;
    return(0);
  }

  /* Will be logging to STDOUT */
  snd_output_stdio_attach(&Log,stdout,0);

  /* Allocate parameters structures */
  if(snd_pcm_hw_params_malloc(&HWParams)<0) return(0);
  if(snd_pcm_sw_params_malloc(&SWParams)<0)
  { snd_pcm_hw_params_free(HWParams);return(0); }

  /* Set hardware parameters */
  J = snd_pcm_hw_params_any(ALSA,HWParams);
  J = J<0? J:snd_pcm_hw_params_set_access(ALSA,HWParams,SNDRV_PCM_ACCESS_RW_INTERLEAVED);
  J = J<0? J:snd_pcm_hw_params_set_format(ALSA,HWParams,sizeof(sample)>1? SNDRV_PCM_FORMAT_S16_LE:SNDRV_PCM_FORMAT_S8);
  J = J<0? J:snd_pcm_hw_params_set_channels(ALSA,HWParams,1);
  J = J<0? J:snd_pcm_hw_params_set_rate_near(ALSA,HWParams,&Rate,0);
  J = J<0? J:snd_pcm_hw_params_set_buffer_size(ALSA,HWParams,SndSize=((Rate*Latency)/1000+31)&~31);
  J = J<0? J:snd_pcm_hw_params_set_periods(ALSA,HWParams,4,0);
  J = J<0? J:snd_pcm_hw_params(ALSA,HWParams);

  /* Set software parameters */
  J = snd_pcm_sw_params_current(ALSA,SWParams);
  J = J<0? J:snd_pcm_sw_params_get_boundary(SWParams,&I);
  J = J<0? J:snd_pcm_sw_params_set_stop_threshold(ALSA,SWParams,I);
  J = J<0? J:snd_pcm_sw_params(ALSA,SWParams);

  /* Done with parameters structures */
  snd_pcm_hw_params_dump(HWParams,Log);
  snd_pcm_sw_params_dump(SWParams,Log);
  snd_pcm_hw_params_free(HWParams);
  snd_pcm_sw_params_free(SWParams);

  /* Sound initialized */
  return(SndRate=J<0? 0:Rate);  
}

/** TrashAudio() *********************************************/
/** Free resources allocated by InitAudio().                **/
/*************************************************************/
void TrashAudio(void)
{
  /* Uninitialize ALSA sound */
  if(ALSA) snd_pcm_close(ALSA);
  ALSA        = 0;
  AudioPaused = 0;
  SndRate     = 0;
}

/** GetFreeAudio() *******************************************/
/** Get the amount of free samples in the audio buffer.     **/
/*************************************************************/
unsigned int GetFreeAudio(void)
{
  int J;
  
  /* Audio should be initialized */
  if(!ALSA) return(0);

  /* Get available frame count */
  J=snd_pcm_avail_update(ALSA);

  /* If failed, try recovering */
  if(J<0) J=snd_pcm_recover(ALSA,J,0);

  /* Report problems */
  if(J<0) printf("ALSA: %s (%d)\n",snd_strerror(J),J);

  /* Done */
  return(J>0? J:0);
}

/** WriteAudio() *********************************************/
/** Write up to a given number of samples to audio buffer.  **/
/** Returns the number of samples written.                  **/
/*************************************************************/
unsigned int WriteAudio(sample *Data,unsigned int Length)
{
  int J;

  /* Audio should be initialized */
  if(!ALSA) return(0);

  /* Send audio data */
  J=snd_pcm_writei(ALSA,Data,Length);

  /* If failed, try recovering */
  if(J<0) J=snd_pcm_recover(ALSA,J,0);

  /* Report problems */
  if(J<0) printf("ALSA: %s (%d)\n",snd_strerror(J),J);

  /* Done */
  return(J>0? J:0);
}

/** PauseAudio() *********************************************/
/** Pause/resume audio playback. Returns current playback   **/
/** state.                                                  **/
/*************************************************************/
int PauseAudio(int Switch)
{
  static int Rate,Latency;

  /* If audio not initialized, return "off" */
  if(!ALSA&&!AudioPaused) return(0);

  /* Toggle audio status if requested */
  if(Switch==2) Switch=AudioPaused? 0:1;

  /* When switching audio state... */
  if((Switch>=0)&&(Switch<=1)&&(Switch!=AudioPaused))
  {
    if(Switch)
    {
      /* Memorize audio parameters and kill audio */
      Rate    = SndRate;
      Latency = 1000*SndSize/SndRate;
      TrashAudio();
    }
    else
    {
      /* Start audio using memorized parameters */
      if(!InitAudio(Rate,Latency)) Switch=0;
    }

    /* Audio switched */
    AudioPaused=Switch;
  }

  /* Return current status */
  return(AudioPaused);
}

/** ALSAVolume() *********************************************/
/** Change ALSA mixer control by a given percentage.        **/
/** Returns the number of samples written.                  **/
/*************************************************************/
#ifndef OLD_ALSA
int ALSAVolume(const char *Name,int Percent)
{
  struct snd_ctl_elem_id ID;
  struct snd_mixer_elem *E;
  struct snd_mixer *H;
  int J,Min,Max,Cur;

  /* Open control and load its data */
  H = 0;
  J = snd_mixer_open(&H,0);
  J = J<0? J:snd_mixer_attach(H,"default");
  J = J<0? J:snd_mixer_selem_register(H,0,0);
  J = J<0? J:snd_mixer_load(H);

  /* Initialize mixer control ID */
  snd_mixer_selem_id_set_index(&ID,0);
  snd_mixer_selem_id_set_name(&ID,Name);

  /* Find control element */
  E = J<0? 0:snd_mixer_find_selem(H,&ID);
  J = E? J:-1;

  /* Get control info */
  Min = 0;
  Max = 100;
  Cur = 50;
  J   = J<0? J:snd_mixer_selem_get_playback_volume_range(E,&Min,&Max);
  J   = J<0? J:snd_mixer_selem_get_playback_volume(E,0,&Cur);

  /* Set control value */
  Cur+= Percent*(Max-Min)/100;
  Cur = Cur<Min? Min:Cur>Max? Max:Cur;
  J   = J<0? J:snd_mixer_selem_set_playback_volume(E,0,Cur);
  J   = J<0? J:snd_mixer_selem_set_playback_volume(E,1,Cur);

  /* Done with control */
  if(H) snd_mixer_close(H);
  return(J<0? -1:Cur);
}
#endif /* !OLD_ALSA */

